@@ -5,8 +5,8 @@ import copy
 import json
 import torch
 import numpy as np
-from sklearn.metrics import accuracy_score
 import backwardcompatibilityml.scores as scores
+from backwardcompatibilityml.metrics import model_accuracy
 
 
 def train_epoch(epoch, network, optimizer, loss_function, training_set, batch_size_train,
@@ -22,7 +22,7 @@ def train_epoch(epoch, network, optimizer, loss_function, training_set, batch_si
         network: The model which is undergoing training.
         optimizer: The optimizer instance to use for training.
         loss_function: An instance of the loss function to use for training.
-        training_set: The list of training samples as (input, target) pairs.
+        training_set: The list of training samples as (batch_ids, input, target).
         batch_size_train: An integer representing batch size of the training set.
         device: A string with values either "cpu" or "cuda" to indicate the
             device that Pytorch is performing training on. By default this
@@ -37,7 +37,7 @@ def train_epoch(epoch, network, optimizer, loss_function, training_set, batch_si
     train_losses = []
     train_counter = []
     network.train()
-    for batch_idx, (data, target) in enumerate(training_set):
+    for batch_idx, (batch_ids, data, target) in enumerate(training_set):
         if device != "cpu":
             data = data.to(device)
             target = target.to(device)
@@ -67,7 +67,7 @@ def test(network, loss_function, test_set, batch_size_test, device="cpu"):
     Args:
         network: The model which is undergoing testing.
         loss_function: An instance of the loss function to use for training.
-        test_set: The list of testing samples as (input, target) pairs.
+        test_set: The list of testing samples as (batch_ids, input, target).
         batch_size_test: An integer representing the batch size of the test set.
         device: A string with values either "cpu" or "cuda" to indicate the
             device that Pytorch is performing training on. By default this
@@ -82,7 +82,7 @@ def test(network, loss_function, test_set, batch_size_test, device="cpu"):
     test_loss = 0
     correct = 0
     with torch.no_grad():
-        for data, target in test_set:
+        for batch_ids, data, target in test_set:
             if device != "cpu":
                 data = data.to(device)
                 target = target.to(device)
@@ -114,8 +114,8 @@ def train(number_of_epochs, network, optimizer, loss_function,
         number_of_epochs: Number of epochs of training.
         optimizer: The optimizer instance to use for training.
         loss_function: An instance of the loss function to use for training.
-        training_set: The list of training samples as (input, target) pairs.
-        test_set: The list of testing samples as (input, target) pairs.
+        training_set: The list of training samples as (batch_ids, input, target).
+        test_set: The list of testing samples as (batch_ids, input, target).
         batch_size_train: An integer representing batch size of the training set.
         batch_size_test: An integer representing the batch size of the test set.
         device: A string with values either "cpu" or "cuda" to indicate the
@@ -162,7 +162,7 @@ def train_compatibility_epoch(epoch, h2, optimizer, loss_function, training_set,
         h2: The model which is undergoing training / updating.
         optimizer: The optimizer instance to use for training.
         loss_function: An instance of a compatibility loss function.
-        training_set: The list of training samples as (input, target) pairs.
+        training_set: The list of training samples as (batch_ids, input, target).
         batch_size_train: An integer representing batch size of the training set.
         device: A string with values either "cpu" or "cuda" to indicate the
             device that Pytorch is performing training on. By default this
@@ -177,7 +177,7 @@ def train_compatibility_epoch(epoch, h2, optimizer, loss_function, training_set,
     train_losses = []
     train_counter = []
     h2.train()
-    for batch_idx, (data, target) in enumerate(training_set):
+    for batch_idx, (batch_ids, data, target) in enumerate(training_set):
         if device != "cpu":
             data = data.to(device)
             target = target.to(device)
@@ -203,7 +203,7 @@ def test_compatibility(h2, loss_function, test_set, batch_size_test, device="cpu
     Args:
         h2: The model which is undergoing training / updating.
         loss_function: An instance of a compatibility loss function.
-        test_set: The list of testing samples as (input, target) pairs.
+        test_set: The list of testing samples as (batch_ids, input, target).
         batch_size_test: An integer representing the batch size of the test set.
         device: A string with values either "cpu" or "cuda" to indicate the
             device that Pytorch is performing training on. By default this
@@ -218,7 +218,7 @@ def test_compatibility(h2, loss_function, test_set, batch_size_test, device="cpu
     test_loss = 0
     correct = 0
     with torch.no_grad():
-        for data, target in test_set:
+        for batch_ids, data, target in test_set:
             if device != "cpu":
                 data = data.to(device)
                 target = target.to(device)
@@ -248,8 +248,8 @@ def train_compatibility(number_of_epochs, h2, optimizer, loss_function,
         number_of_epochs: Number of epochs of training.
         loss_function: An instance of a compatibility loss function.
         optimizer: The optimizer instance to use for training.
-        training_set: The list of training samples as (input, target) pairs.
-        test_set: The list of testing samples as (input, target) pairs.
+        training_set: The list of training samples as (batch_ids, input, target).
+        test_set: The list of testing samples as (batch_ids, input, target).
         batch_size_train: An integer representing batch size of the training set.
         batch_size_test: An integer representing the batch size of the test set.
         device: A string with values either "cpu" or "cuda" to indicate the
@@ -313,15 +313,98 @@ def get_error_instance_indices(model, batched_evaluation_data, batched_evaluatio
         return model_diff.nonzero().view(-1).tolist()
 
 
-def get_normalized_model_error_overlap(h1, h2, batched_evaluation_data, batched_evaluation_target,
-                                       device="cpu"):
+def get_all_error_instance_indices(h1, h2, batch_ids, batched_evaluation_data, batched_evaluation_target,
+                                   get_instance_metadata=None,
+                                   device="cpu"):
     """
-    Return the fraction of errors of each model and the fraction of errors common to both models,
-    all with respect to the total number of errors of each model.
+    Return the list of indices of instances from batched_evaluation_data on which the
+    model prediction differs from the ground truth in batched_evaluation_target.
+
+    Args:
+        h1: The baseline model.
+        h2: The new updated model.
+        batch_ids: A list of the instance ids in the batch.
+        batched_evaluation_data: A single batch of input data to be passed to our model.
+        batched_evaluation_target: A single batch of the corresponding output targets.
+        get_instance_metadata: A function that returns a text string representation of some metadata corresponding to the instance id. It should be a function of the form: 
+                get_instance_metadata(instance_id)
+                    instance_id: An integer instance id
+
+            And should return a string.
+        device: A string with values either "cpu" or "cuda" to indicate the
+            device that Pytorch is performing training on. By default this
+            value is "cpu". But in case your models reside on the GPU, make sure
+            to set this to "cuda". This makes sure that the input and target
+            tensors are transferred to the GPU during training.
+    Returns:
+        A list of indices of the instances within the batched data, for which the
+        model did not match the expected target.
+    """
+    with torch.no_grad():
+        if device != "cpu":
+            batched_evaluation_data = batched_evaluation_data.to(device)
+            batched_evaluation_target = batched_evaluation_target.to(device)
+        _, _, h1_output_logsoftmax = h1(batched_evaluation_data)
+        _, _, h2_output_logsoftmax = h2(batched_evaluation_data)
+        h1_diff = (torch.argmax(h1_output_logsoftmax, 1) - batched_evaluation_target).float()
+        h2_diff = (torch.argmax(h2_output_logsoftmax, 1) - batched_evaluation_target).float()
+        h1_error_instance_ids = h1_diff.nonzero().view(-1).tolist()
+        h2_error_instance_ids = h2_diff.nonzero().view(-1).tolist()
+        error_instance_ids = list(set(h1_error_instance_ids).union(set(h2_error_instance_ids)))
+        error_instances_metadata = [""] * len(error_instance_ids)
+        if get_instance_metadata is not None:
+            error_instances_metadata = list(map(get_instance_metadata, error_instance_ids))
+        error_instance_batch_ids = []
+        h1_predictions = []
+        h2_predictions = []
+        instance_ground_truths = []
+        if len(error_instance_ids) > 0:
+            error_instance_batch_ids = torch.tensor(batch_ids).index_select(
+                0, torch.tensor(error_instance_ids)).tolist()
+            h1_predictions = torch.argmax(h1_output_logsoftmax, 1).index_select(
+                0, torch.tensor(error_instance_ids).to(device)).tolist()
+            h2_predictions = torch.argmax(h2_output_logsoftmax, 1).index_select(
+                0, torch.tensor(error_instance_ids).to(device)).tolist()
+            instance_ground_truths = batched_evaluation_target.index_select(
+                0, torch.tensor(error_instance_ids).to(device)).tolist()
+
+        return list(zip(error_instance_batch_ids,
+                    error_instances_metadata,
+                    h1_predictions,
+                    h2_predictions,
+                    instance_ground_truths))
+
+
+def get_incompatible_instances_by_class(all_errors, batch_ids, batched_evaluation_target, class_incompatible_instance_ids):
+    """
+    Finds instances where h2 is incompatible with h1 and inserts
+    {class : incompatible_data_id} mappings into the class_incompatible_instance_ids dictionary.
+
+    Args:
+        all_errors: A list of tuples of error indices, h1 and h2 predictions, and ground truth for each instance
+        batch_ids: The instance ids of the data rows in the batched data.
+        batched_evaluation_target: A single batch of the corresponding output targets.
+        class_incompatible_instance_ids: The dictionary to fill with incompatible instances and their ids
+    """
+    for (error_instance_id, error_instance_metadata, h1_prediction, h2_prediction, ground_truth) in all_errors:
+        if (h1_prediction == ground_truth and h2_prediction != ground_truth):
+            batch_index = batch_ids.index(error_instance_id)
+            incompatible_class = batched_evaluation_target[batch_index].item()
+            if (incompatible_class not in class_incompatible_instance_ids):
+                class_incompatible_instance_ids[incompatible_class] = []
+            class_incompatible_instance_ids[incompatible_class].append(error_instance_id)
+
+
+def get_model_error_overlap(h1, h2, batch_ids, batched_evaluation_data, batched_evaluation_target,
+                            device="cpu"):
+    """
+    Return the instance ids corresponding to errors of each model
+    as well as the instance ids corresponding to errors common to both models.
 
     Args:
         h1: Reference Pytorch model.
         h2: The model being compared to h1.
+        batch_ids: The instance ids of the data rows in the batched data.
         batched_evaluation_data: A single batch of input data to be passed to our model.
         batched_evaluation_target: A single batch of the corresponding output targets.
         device: A string with values either "cpu" or "cuda" to indicate the
@@ -330,32 +413,39 @@ def get_normalized_model_error_overlap(h1, h2, batched_evaluation_data, batched_
             to set this to "cuda". This makes sure that the input and target
             tensors are transferred to the GPU during training.
     Returns:
-        |  If there are any errors at all it returns a triple of the form 
-            proportion_of_error_due_to_h1, 
-            proportion_of_error_due_to_h2, 
-            proportion_of_error_due_to_h1_and_h2
-        |  If there are no errors at all it returns the triple 
-            0, 0, 0
+        A triple of the form:
+            instance_ids_of_errors_due_to_h1,
+            instance_ids_of_errors_due_to_h2,
+            instance_ids_of_errors_due_to_h1_and_h2
     """
     h1_error_indices = get_error_instance_indices(
         h1, batched_evaluation_data, batched_evaluation_target, device=device)
     h2_error_indices = get_error_instance_indices(
         h2, batched_evaluation_data, batched_evaluation_target, device=device)
-    h1_size = len(h1_error_indices)
-    h2_size = len(h2_error_indices)
-    h1_and_h2_size = len(set(h1_error_indices).intersection(set(h2_error_indices)))
-    total = h1_size + h2_size - h1_and_h2_size
+    h1_and_h2_error_indices = list(set(h1_error_indices).intersection(set(h2_error_indices)))
 
-    if total > 0:
-        return (h1_size / total), (h2_size / total), (h1_and_h2_size / total)
+    h1_error_instance_ids = []
+    h2_error_instance_ids = []
+    h1_and_h2_error_instance_ids = []
+    if len(h1_error_indices) > 0:
+        h1_error_instance_ids = torch.tensor(batch_ids).index_select(
+            0, torch.tensor(h1_error_indices)).tolist()
 
-    return 0, 0, 0
+    if len(h2_error_indices) > 0:
+        h2_error_instance_ids = torch.tensor(batch_ids).index_select(
+            0, torch.tensor(h2_error_indices)).tolist()
+
+    if len(h1_and_h2_error_indices) > 0:
+        h1_and_h2_error_instance_ids = torch.tensor(batch_ids).index_select(
+            0, torch.tensor(h1_and_h2_error_indices)).tolist()
+
+    return h1_error_instance_ids, h2_error_instance_ids, h1_and_h2_error_instance_ids
 
 
-def get_error_fraction_by_class(model, batched_evaluation_data, batched_evaluation_target,
-                                device="cpu"):
+def get_error_instance_ids_by_class(model, batch_ids, batched_evaluation_data, batched_evaluation_target,
+                                    device="cpu"):
     """
-    Return the fraction of errors of the model by class.
+    Return the instance ids corresponding to errors of the model by class.
 
     Args:
         model: The model being evaluated.
@@ -368,29 +458,27 @@ def get_error_fraction_by_class(model, batched_evaluation_data, batched_evaluati
             tensors are transferred to the GPU during training.
     Returns:
         A dictionary of key / value pairs, where the key is the output class
-        and the value is the fraction of misclassification errors of the
-        model within that class.
+        and the value is the list of instance ids corresponding to misclassification
+        errors of the model within that class.
     """
     with torch.no_grad():
         if device != "cpu":
             batched_evaluation_data = batched_evaluation_data.to(device)
             batched_evaluation_target = batched_evaluation_target.to(device)
         _, _, model_output_logsoftmax = model(batched_evaluation_data)
+        class_error_instance_ids = {}
         model_diff = (torch.argmax(model_output_logsoftmax, 1) - batched_evaluation_target).float()
-        model_errors = (model_diff != 0)
-        target_error_classes = batched_evaluation_target[model_errors].view(-1).tolist()
-        class_error_count = {}
-        total_errors = len(target_error_classes)
-        for class_label in set(batched_evaluation_target.view(-1).tolist()):
-            class_error_count[class_label] = 0
-        for error_class in target_error_classes:
-            class_error_count[error_class] = class_error_count[error_class] + 1
+        model_errors = model_diff.nonzero().view(-1).tolist()
+        if len(model_errors) > 0:
+            error_instance_indices = torch.tensor(batch_ids).index_select(
+                0, torch.tensor(model_errors)).tolist()
+            target_error_classes = batched_evaluation_target[model_errors].view(-1).tolist()
+            for class_label in set(batched_evaluation_target.view(-1).tolist()):
+                class_error_instance_ids[class_label] = []
+            for error_instance_index, error_class in zip(error_instance_indices, target_error_classes):
+                class_error_instance_ids[error_class].append(error_instance_index)
 
-        if total_errors > 0:
-            for class_label, error_count in class_error_count.items():
-                class_error_count[class_label] = error_count / total_errors
-
-        return class_error_count
+        return class_error_instance_ids
 
 
 def compatibility_scores(h1, h2, dataset, device="cpu"):
@@ -413,7 +501,7 @@ def compatibility_scores(h1, h2, dataset, device="cpu"):
     with torch.no_grad():
         btc_dataset = 0
         bec_dataset = 0
-        for data, target in dataset:
+        for batch_ids, data, target in dataset:
             if device != "cpu":
                 data = data.to(device)
                 target = target.to(device)
@@ -432,36 +520,19 @@ def compatibility_scores(h1, h2, dataset, device="cpu"):
     return btc_dataset, bec_dataset
 
 
-def model_accuracy(model, dataset, device="cpu"):
-    number_of_batches = len(dataset)
-    model_performance = 0
-    with torch.no_grad():
-        for data, target in dataset:
-            if device != "cpu":
-                data = data.to(device)
-                target = target.to(device)
-            _, _, output_logsoftmax = model(data)
-            output_labels = torch.argmax(output_logsoftmax, 1)
-            if device != "cpu":
-                output_labels = output_labels.cpu()
-                target = target.cpu()
-            performance = accuracy_score(output_labels.numpy(), target.numpy())
-            model_performance += performance
-            # _clean_from_gpu([data, target])
-
-        model_performance /= number_of_batches
-
-    return model_performance
-
-
-def evaluate_model_performance_and_compatibility_on_dataset(h1, h2, dataset, performance_metric=None,
+def evaluate_model_performance_and_compatibility_on_dataset(h1, h2, dataset, performance_metric,
+                                                            get_instance_metadata=None,
                                                             device="cpu"):
     """
     Args:
         h1: The reference model being used.
         h2: The model being traind / updated.
-        performance_metric: Optional performance metric to be used when evaluating the model.
-            If not specified then accuracy is used.
+        performance_metric: Performance metric to be used when evaluating the model.
+        get_instance_metadata: A function that returns a text string representation of some metadata corresponding to the instance id. It should be a function of the form:
+                get_instance_metadata(instance_id)
+                    instance_id: An integer instance id
+                    
+            And should return a string.
         device: A string with values either "cpu" or "cuda" to indicate the
             device that Pytorch is performing training on. By default this
             value is "cpu". But in case your models reside on the GPU, make sure
@@ -474,77 +545,66 @@ def evaluate_model_performance_and_compatibility_on_dataset(h1, h2, dataset, per
         the error compatibility score of h2 with respect to h1.
     """
     number_of_batches = len(dataset)
-    h1_dataset_error_fraction = 0
-    h2_dataset_error_fraction = 0
-    h1_and_h2_dataset_error_fraction = 0
-    h2_dataset_error_fraction_by_class = {}
+    h1_dataset_error_instance_ids = []
+    h2_dataset_error_instance_ids = []
+    h1_and_h2_dataset_error_instance_ids = []
+    h1h2_dataset_incompatible_instance_ids_by_class = {}
     classes = set()
-    for data, target in dataset:
+    all_error_instances = []
+    dataset_size = 0
+    for batch_ids, data, target in dataset:
+        dataset_size += len(batch_ids)
         classes = classes.union(target.tolist())
-        h1_error_fraction, h2_error_fraction, h1_and_h2_error_fraction =\
-            get_normalized_model_error_overlap(h1, h2, data, target, device=device)
-        h2_error_fraction_by_class =\
-            get_error_fraction_by_class(h2, data, target, device=device)
-        h1_dataset_error_fraction += h1_error_fraction
-        h2_dataset_error_fraction += h2_error_fraction
-        h1_and_h2_dataset_error_fraction += h1_and_h2_error_fraction
-        for class_label, error_fraction in h2_error_fraction_by_class.items():
-            if class_label in h2_dataset_error_fraction_by_class:
-                h2_dataset_error_fraction_by_class[class_label] += error_fraction
-            else:
-                h2_dataset_error_fraction_by_class[class_label] = error_fraction
+        h1_error_count_batch, h2_error_count_batch, h1_and_h2_error_count_batch =\
+            get_model_error_overlap(h1, h2, batch_ids, data, target, device=device)
+        all_errors = get_all_error_instance_indices(
+            h1, h2, batch_ids, data, target,
+            get_instance_metadata=get_instance_metadata, device=device)
+        get_incompatible_instances_by_class(all_errors, batch_ids, target, h1h2_dataset_incompatible_instance_ids_by_class)
+        all_error_instances += all_errors
+        h1_dataset_error_instance_ids += h1_error_count_batch
+        h2_dataset_error_instance_ids += h2_error_count_batch
+        h1_and_h2_dataset_error_instance_ids += h1_and_h2_error_count_batch
 
-    h1_dataset_error_fraction /= number_of_batches
-    h2_dataset_error_fraction /= number_of_batches
-    h1_and_h2_dataset_error_fraction /= number_of_batches
-
-    for class_label, error_fraction in h2_dataset_error_fraction_by_class.items():
-        h2_dataset_error_fraction_by_class[class_label] /= number_of_batches
-
-    h2_ds_error_fraction_by_class = []
-    for class_label, error_fraction in h2_dataset_error_fraction_by_class.items():
-        h2_ds_error_fraction_by_class.append({
+    h1h2_ds_incompatible_instance_ids_by_class = []
+    for class_label, incompatible_instance_ids in h1h2_dataset_incompatible_instance_ids_by_class.items():
+        h1h2_ds_incompatible_instance_ids_by_class.append({
             "class": class_label,
-            "incompatibleFraction": error_fraction
+            "incompatibleInstanceIds": incompatible_instance_ids
         })
 
-    if performance_metric is not None:
-        h2_performance = performance_metric(h2, dataset)
-    else:
-        h2_performance = 0
-        with torch.no_grad():
-            for data, target in dataset:
-                if device != "cpu":
-                    data = data.to(device)
-                    target = target.to(device)
-                _, _, output_logsoftmax = h2(data)
-                output_labels = torch.argmax(output_logsoftmax, 1)
-                if device != "cpu":
-                    output_labels = output_labels.cpu()
-                    target = target.cpu()
-                performance = accuracy_score(output_labels.numpy(), target.numpy())
-                h2_performance += performance
-                # _clean_from_gpu([data, target])
+    h2_performance = performance_metric(h2, dataset, device)
 
-            h2_performance /= number_of_batches
+    all_error_instances_results = []
+    for error_instance_id, error_instance_metadata, h1_prediction, h2_prediction, ground_truth in all_error_instances:
+        all_error_instances_results.append({
+            "instance_id": error_instance_id,
+            "metadata": error_instance_metadata,
+            "h1_prediction": h1_prediction,
+            "h2_prediction": h2_prediction,
+            "ground_truth": ground_truth
+        })
 
     btc, bec = compatibility_scores(h1, h2, dataset, device=device)
 
     return {
         "models_error_overlap": [
-            h1_dataset_error_fraction,
-            h2_dataset_error_fraction,
-            h1_and_h2_dataset_error_fraction
+            h1_dataset_error_instance_ids,
+            h2_dataset_error_instance_ids,
+            h1_and_h2_dataset_error_instance_ids
         ],
 
-        "h2_error_fraction_by_class": h2_ds_error_fraction_by_class,
+        "h2_incompatible_instance_ids_by_class": h1h2_ds_incompatible_instance_ids_by_class,
+        "sorted_classes": sorted(list(classes)),
         "h2_performance": h2_performance,
         "btc": btc,
-        "bec": bec
+        "bec": bec,
+        "error_instances": all_error_instances_results,
+        "dataset_size": dataset_size
     }
 
 
-def evaluate_model_performance_and_compatibility(h1, h2, training_set, test_set, performance_metric=None,
+def evaluate_model_performance_and_compatibility(h1, h2, training_set, test_set, performance_metric,
                                                  device="cpu"):
     """
     Calculate the error overlap of h1 and h2 on a batched dataset.
@@ -553,10 +613,9 @@ def evaluate_model_performance_and_compatibility(h1, h2, training_set, test_set,
     Args:
         h1: The reference model being used.
         h2: The model being traind / updated.
-        performance_metric: Optional performance metric to be used when evaluating the model.
-            If not specified then accuracy is used.
-        training_set: The list of batched training samples as (input, target) pairs.
-        test_set: The list of batched testing samples as (input, target) pairs.
+        performance_metric: Performance metric to be used when evaluating the model.
+        training_set: The list of batched training samples as (batch_ids, input, target).
+        test_set: The list of batched testing samples as (batch_ids, input, target).
         device: A string with values either "cpu" or "cuda" to indicate the
             device that Pytorch is performing training on. By default this
             value is "cpu". But in case your models reside on the GPU, make sure
@@ -568,11 +627,11 @@ def evaluate_model_performance_and_compatibility(h1, h2, training_set, test_set,
     """
     training_set_performance_and_compatibility =\
         evaluate_model_performance_and_compatibility_on_dataset(
-            h1, h2, training_set, performance_metric=performance_metric,
+            h1, h2, training_set, performance_metric,
             device=device)
     testing_set_performance_and_compatibility =\
         evaluate_model_performance_and_compatibility_on_dataset(
-            h1, h2, test_set, performance_metric=performance_metric,
+            h1, h2, test_set, performance_metric,
             device=device)
 
     return {
@@ -591,8 +650,8 @@ def train_new_error(h1, h2, number_of_epochs,
         h1: Reference Pytorch model.
         h2: The model which is undergoing training / updating.
         number_of_epochs: Number of epochs of training.
-        training_set: The list of training samples as (input, target) pairs.
-        test_set: The list of testing samples as (input, target) pairs.
+        training_set: The list of training samples as (batch_ids, input, target).
+        test_set: The list of testing samples as (batch_ids, input, target).
         batch_size_train: An integer representing batch size of the training set.
         batch_size_test: An integer representing the batch size of the test set.
         OptimizerClass: The class to instantiate an optimizer from for training.
@@ -626,8 +685,8 @@ def train_strict_imitation(h1, h2, number_of_epochs,
         h1: Reference Pytorch model.
         h2: The model which is undergoing training / updating.
         number_of_epochs: Number of epochs of training.
-        training_set: The list of training samples as (input, target) pairs.
-        test_set: The list of testing samples as (input, target) pairs.
+        training_set: The list of training samples as (batch_ids, input, target).
+        test_set: The list of testing samples as (batch_ids, input, target).
         batch_size_train: An integer representing batch size of the training set.
         batch_size_test: An integer representing the batch size of the test set.
         OptimizerClass: The class to instantiate an optimizer from for training.
@@ -655,10 +714,11 @@ def compatibility_sweep(sweeps_folder_path, number_of_epochs, h1, h2,
                         training_set, test_set, batch_size_train, batch_size_test,
                         OptimizerClass, optimizer_kwargs,
                         NewErrorLossClass, StrictImitationLossClass,
-                        performance_metric=None,
+                        performance_metric=model_accuracy,
                         lambda_c_stepsize=0.25, percent_complete_queue=None,
                         new_error_loss_kwargs=None,
                         strict_imitation_loss_kwargs=None,
+                        get_instance_metadata=None,
                         device="cpu"):
     """
     This function trains a new model using the backward compatibility loss function
@@ -672,8 +732,8 @@ def compatibility_sweep(sweeps_folder_path, number_of_epochs, h1, h2,
         number_of_epochs: The number of training epochs to use on each sweep.
         h1: The reference model being used.
         h2: The new model being traind / updated.
-        training_set: The list of training samples as (input, target) pairs.
-        test_set: The list of testing samples as (input, target) pairs.
+        training_set: The list of training samples as (batch_ids, input, target).
+        test_set: The list of testing samples as (batch_ids, input, target).
         batch_size_train: An integer representing batch size of the training set.
         batch_size_test: An integer representing the batch size of the test set.
         OptimizerClass: The class to instantiate an optimizer from for training.
@@ -685,12 +745,22 @@ def compatibility_sweep(sweeps_folder_path, number_of_epochs, h1, h2,
         StrictImitationLossClass: The class of the Strict Imitation style loss
             function to be instantiated and used to perform compatibility
             constrained training of our model h2.
-        performance_metric: Optional performance metric to be used when evaluating the model.
-            If not specified then accuracy is used.
+        performance_metric: A function to evaluate model performance. The function is expected to have the following signature:
+                metric(model, dataset, device)
+                    model: The model being evaluated
+                    dataset: The dataset as a list of (batch_ids, input, target)
+                    device: The device Pytorch is using for training - "cpu" or "cuda"
+
+            If unspecified, then accuracy is used.
         lambda_c_stepsize: The increments of lambda_c to use as we sweep the parameter
             space between 0.0 and 1.0.
         percent_complete_queue: Optional thread safe queue to use for logging the
             status of the sweep in terms of the percentage complete.
+        get_instance_metadata: A function that returns a text string representation of some metadata corresponding to the instance id. It should be a function of the form:
+                get_instance_metadata(instance_id)
+                    instance_id: An integer instance id
+
+            And should return a string.
         device: A string with values either "cpu" or "cuda" to indicate the
             device that Pytorch is performing training on. By default this
             value is "cpu". But in case your models reside on the GPU, make sure
@@ -723,7 +793,8 @@ def compatibility_sweep(sweeps_folder_path, number_of_epochs, h1, h2,
 
         training_set_performance_and_compatibility =\
             evaluate_model_performance_and_compatibility_on_dataset(
-                h1, h2_new_error, training_set, performance_metric=performance_metric,
+                h1, h2_new_error, training_set, performance_metric,
+                get_instance_metadata=get_instance_metadata,
                 device=device)
         training_set_performance_and_compatibility["lambda_c"] = lambda_c
         training_set_performance_and_compatibility["training"] = True
@@ -748,7 +819,8 @@ def compatibility_sweep(sweeps_folder_path, number_of_epochs, h1, h2,
 
         testing_set_performance_and_compatibility =\
             evaluate_model_performance_and_compatibility_on_dataset(
-                h1, h2_new_error, test_set, performance_metric=performance_metric,
+                h1, h2_new_error, test_set, performance_metric,
+                get_instance_metadata=get_instance_metadata,
                 device=device)
         testing_set_performance_and_compatibility["lambda_c"] = lambda_c
         testing_set_performance_and_compatibility["training"] = False
@@ -783,7 +855,8 @@ def compatibility_sweep(sweeps_folder_path, number_of_epochs, h1, h2,
 
         training_set_performance_and_compatibility =\
             evaluate_model_performance_and_compatibility_on_dataset(
-                h1, h2_strict_imitation, training_set, performance_metric=performance_metric,
+                h1, h2_strict_imitation, training_set, performance_metric,
+                get_instance_metadata=get_instance_metadata,
                 device=device)
         training_set_performance_and_compatibility["lambda_c"] = lambda_c
         training_set_performance_and_compatibility["training"] = True
@@ -808,7 +881,8 @@ def compatibility_sweep(sweeps_folder_path, number_of_epochs, h1, h2,
 
         testing_set_performance_and_compatibility =\
             evaluate_model_performance_and_compatibility_on_dataset(
-                h1, h2_new_error, test_set, performance_metric=performance_metric,
+                h1, h2_new_error, test_set, performance_metric,
+                get_instance_metadata=get_instance_metadata,
                 device=device)
         testing_set_performance_and_compatibility["lambda_c"] = lambda_c
         testing_set_performance_and_compatibility["training"] = False
@@ -836,7 +910,7 @@ def compatibility_sweep(sweeps_folder_path, number_of_epochs, h1, h2,
 
     sweep_summary = {
         "data": sweep_summary_data,
-        "h1_performance": model_accuracy(h1, test_set)
+        "h1_performance": model_accuracy(h1, test_set, device=device)
     }
 
     sweep_summary_data = json.dumps(sweep_summary)
