@@ -4,8 +4,13 @@
 import os
 import json
 import threading
+import io
+import numpy as np
+from flask import send_file
+from PIL import Image
 from queue import Queue
 from backwardcompatibilityml.helpers import training
+from backwardcompatibilityml.metrics import model_accuracy
 
 
 class SweepManager(object):
@@ -28,8 +33,8 @@ class SweepManager(object):
         number_of_epochs: The number of training epochs to use on each sweep.
         h1: The reference model being used.
         h2: The new model being traind / updated.
-        training_set: The list of training samples as (input, target) pairs.
-        test_set: The list of testing samples as (input, target) pairs.
+        training_set: The list of training samples as (batch_ids, input, target).
+        test_set: The list of testing samples as (batch_ids, input, target).
         batch_size_train: An integer representing batch size of the training set.
         batch_size_test: An integer representing the batch size of the test set.
         OptimizerClass: The class to instantiate an optimizer from for training.
@@ -45,6 +50,16 @@ class SweepManager(object):
             If not specified then accuracy is used.
         lambda_c_stepsize: The increments of lambda_c to use as we sweep the parameter
             space between 0.0 and 1.0.
+        get_instance_image_by_id: A function that returns an image representation of the data corresponding to the instance id, in PNG format. It should be a function of the form:
+                get_instance_image_by_id(instance_id)
+                    instance_id: An integer instance id
+
+            And should return a PNG image.
+        get_instance_metadata: A function that returns a text string representation of some metadata corresponding to the instance id. It should be a function of the form:
+                get_instance_metadata(instance_id)
+                    instance_id: An integer instance id
+
+            And should return a string.
         device: A string with values either "cpu" or "cuda" to indicate the
             device that Pytorch is performing training on. By default this
             value is "cpu". But in case your models reside on the GPU, make sure
@@ -58,6 +73,9 @@ class SweepManager(object):
                  NewErrorLossClass, StrictImitationLossClass, lambda_c_stepsize=0.25,
                  new_error_loss_kwargs=None,
                  strict_imitation_loss_kwargs=None,
+                 performance_metric=model_accuracy,
+                 get_instance_image_by_id=None,
+                 get_instance_metadata=None,
                  device="cpu"):
         self.folder_name = folder_name
         self.number_of_epochs = number_of_epochs
@@ -71,9 +89,12 @@ class SweepManager(object):
         self.optimizer_kwargs = optimizer_kwargs
         self.NewErrorLossClass = NewErrorLossClass
         self.StrictImitationLossClass = StrictImitationLossClass
+        self.performance_metric = performance_metric
         self.lambda_c_stepsize = lambda_c_stepsize
         self.new_error_loss_kwargs = new_error_loss_kwargs
         self.strict_imitation_loss_kwargs = strict_imitation_loss_kwargs
+        self.get_instance_image_by_id = get_instance_image_by_id
+        self.get_instance_metadata = get_instance_metadata
         self.device = device
         self.last_sweep_status = 0.0
         self.percent_complete_queue = Queue()
@@ -83,12 +104,14 @@ class SweepManager(object):
                   self.training_set, self.test_set,
                   self.batch_size_train, self.batch_size_test,
                   self.OptimizerClass, self.optimizer_kwargs,
-                  self.NewErrorLossClass, self.StrictImitationLossClass,),
+                  self.NewErrorLossClass, self.StrictImitationLossClass,
+                  self.performance_metric,),
             kwargs={
                 "lambda_c_stepsize": self.lambda_c_stepsize,
                 "percent_complete_queue": self.percent_complete_queue,
                 "new_error_loss_kwargs": self.new_error_loss_kwargs,
                 "strict_imitation_loss_kwargs": self.strict_imitation_loss_kwargs,
+                "get_instance_metadata": self.get_instance_metadata,
                 "device": self.device
             })
 
@@ -104,6 +127,7 @@ class SweepManager(object):
             lambda_c_stepsize=self.lambda_c_stepsize, percent_complete_queue=self.percent_complete_queue,
             new_error_loss_kwargs=self.new_error_loss_kwargs,
             strict_imitation_loss_kwargs=self.strict_imitation_loss_kwargs,
+            get_instance_metadata=self.get_instance_metadata,
             device=self.device)
 
     def get_sweep_status(self):
@@ -116,6 +140,7 @@ class SweepManager(object):
     def get_sweep_summary(self):
         sweep_summary = {
             "h1_performance": None,
+            "performance_metric": self.performance_metric.__name__,
             "data": []
         }
 
@@ -131,3 +156,17 @@ class SweepManager(object):
             evaluation_data = json.loads(evaluation_data_file.read())
 
         return evaluation_data
+
+    def get_instance_image(self, instance_id):
+        get_instance_image_by_id = self.get_instance_image_by_id
+        if get_instance_image_by_id is not None:
+            return get_instance_image_by_id(instance_id)
+
+        # Generate a blank white PNG image as the default
+        data = np.uint8(np.zeros((30, 30)) + 255)
+        image = Image.fromarray(data)
+        img_bytes = io.BytesIO()
+        image.save(img_bytes, format="PNG")
+        img_bytes.seek(0)
+
+        return send_file(img_bytes, mimetype="image/png")
